import io
import os
import unittest
from unittest import mock
import subprocess
import tempfile
import multiprocessing
import itertools
import time
import uuid
import typing

from collections import OrderedDict

from autograder_sandbox import AutograderSandbox, VERSION


def kb_to_bytes(num_kb):
    return 1000 * num_kb


def mb_to_bytes(num_mb):
    return 1000 * kb_to_bytes(num_mb)


def gb_to_bytes(num_gb):
    return 1000 * mb_to_bytes(num_gb)


class AutograderSandboxInitTestCase(unittest.TestCase):

    def setUp(self):
        self.name = 'awexome_container{}'.format(uuid.uuid4().hex)
        self.environment_variables = OrderedDict(
            {'spam': 'egg', 'sausage': 42})

    def test_default_init(self):
        sandbox = AutograderSandbox()
        self.assertIsNotNone(sandbox.name)
        self.assertFalse(sandbox.allow_network_access)
        self.assertEqual({}, sandbox.environment_variables)
        self.assertEqual('jameslp/autograder-sandbox:{}'.format(VERSION), sandbox.docker_image)

    def test_non_default_init(self):
        docker_image = 'waaaaluigi'
        sandbox = AutograderSandbox(
            name=self.name,
            docker_image=docker_image,
            allow_network_access=True,
            environment_variables=self.environment_variables
        )
        self.assertEqual(self.name, sandbox.name)
        self.assertEqual(docker_image, sandbox.docker_image)
        self.assertTrue(sandbox.allow_network_access)
        self.assertEqual(self.environment_variables, sandbox.environment_variables)


class AutograderSandboxBasicRunCommandTestCase(unittest.TestCase):

    def setUp(self):
        self.sandbox = AutograderSandbox()

        self.root_cmd = ["touch", "/"]

    def test_run_legal_command_non_root(self):
        stdout_content = "hello world"
        expected_output = stdout_content.encode() + b'\n'
        with self.sandbox:
            cmd_result = self.sandbox.run_command(["echo", stdout_content])
            self.assertEqual(0, cmd_result.return_code)
            self.assertEqual(expected_output, cmd_result.stdout.read())

    def test_run_illegal_command_non_root(self):
        with self.sandbox:
            cmd_result = self.sandbox.run_command(self.root_cmd)
            self.assertNotEqual(0, cmd_result.return_code)
            self.assertNotEqual("", cmd_result.stderr)

    def test_run_command_as_root(self):
        with self.sandbox:
            cmd_result = self.sandbox.run_command(self.root_cmd, as_root=True)
            self.assertEqual(0, cmd_result.return_code)
            self.assertEqual(b"", cmd_result.stderr.read())

    def test_run_command_raise_on_error(self):
        """
        Tests that an exception is thrown only when check is True
        and the command exits with nonzero status.
        """
        with self.sandbox:
            # No exception should be raised.
            cmd_result = self.sandbox.run_command(self.root_cmd, as_root=True, check=True)
            self.assertEqual(0, cmd_result.return_code)

            with self.assertRaises(subprocess.CalledProcessError):
                self.sandbox.run_command(self.root_cmd, check=True)

    def test_run_command_executable_does_not_exist_no_error(self):
        with self.sandbox:
            cmd_result = self.sandbox.run_command(['not_an_exe'])
            self.assertNotEqual(0, cmd_result.return_code)


class AutograderSandboxMiscTestCase(unittest.TestCase):

    def setUp(self):
        self.name = 'awexome_container{}'.format(uuid.uuid4().hex)
        self.environment_variables = OrderedDict(
            {'spam': 'egg', 'sausage': 42})

        self.stdin = tempfile.NamedTemporaryFile()
        self.stdout = tempfile.NamedTemporaryFile()
        self.stderr = tempfile.NamedTemporaryFile()

    def tearDown(self):
        self.stdin.close()
        self.stdout.close()
        self.stderr.close()

    def _write_and_seek(self, file_obj, content):
        file_obj.write(content)
        file_obj.seek(0)

    def test_very_large_io_no_truncate(self):
        repeat_str = b'a' * 1000
        num_repeats = 1000000  # 1 GB
        for i in range(num_repeats):
            self.stdin.write(repeat_str)
        with AutograderSandbox() as sandbox:  # type: AutograderSandbox
            self.stdin.seek(0)
            start = time.time()
            result = sandbox.run_command(['cat'], stdin=self.stdin)
            self.assertFalse(result.stdout_truncated)
            self.assertFalse(result.stderr_truncated)
            print('Ran command that read and printed {} bytes to stdout in {}'.format(
                    num_repeats * len(repeat_str), time.time() - start))
            stdout_size = os.path.getsize(result.stdout.name)
            print(stdout_size)
            self.assertEqual(len(repeat_str) * num_repeats, stdout_size)

        with AutograderSandbox() as sandbox:  # type: AutograderSandbox
            self.stdin.seek(0)
            start = time.time()
            result = sandbox.run_command(['bash', '-c', '>&2 cat'], stdin=self.stdin)
            print('Ran command that read and printed {} bytes to stderr in {}'.format(
                num_repeats * len(repeat_str), time.time() - start))
            stderr_size = os.path.getsize(result.stderr.name)
            print(stderr_size)
            self.assertEqual(len(repeat_str) * num_repeats, stderr_size)

    def test_truncate_stdout(self):
        truncate_length = 10
        long_output = b'a' * 100
        expected_output = long_output[:truncate_length]
        self._write_and_seek(self.stdin, long_output)
        with AutograderSandbox() as sandbox:  # type: AutograderSandbox
            result = sandbox.run_command(
                ['cat'], stdin=self.stdin, truncate_stdout=truncate_length)
            self.assertEqual(expected_output, result.stdout.read())
            self.assertTrue(result.stdout_truncated)
            self.assertFalse(result.stderr_truncated)

    def test_truncate_stderr(self):
        truncate_length = 10
        long_output = b'a' * 100
        expected_output = long_output[:truncate_length]
        self._write_and_seek(self.stdin, long_output)
        with AutograderSandbox() as sandbox:  # type: AutograderSandbox
            result = sandbox.run_command(
                ['bash', '-c', '>&2 cat'], stdin=self.stdin, truncate_stderr=truncate_length)
            self.assertEqual(expected_output, result.stderr.read())
            self.assertTrue(result.stderr_truncated)
            self.assertFalse(result.stdout_truncated)

    def test_run_command_with_input(self):
        expected_stdout = b'spam egg sausage spam'
        self._write_and_seek(self.stdin, expected_stdout)
        with AutograderSandbox() as sandbox:  # type: AutograderSandbox
            result = sandbox.run_command(['cat'], stdin=self.stdin)
            self.assertEqual(expected_stdout, result.stdout.read())

    def test_return_code_reported_and_stderr_recorded(self):
        with AutograderSandbox() as sandbox:
            result = sandbox.run_command(['ls', 'definitely not a file'])
            self.assertNotEqual(0, result.return_code)
            self.assertNotEqual('', result.stderr)

    def test_context_manager(self):
        with AutograderSandbox(name=self.name) as sandbox:  # type: AutograderSandbox
            self.assertEqual(self.name, sandbox.name)
            # If the container was created successfully, we
            # should get an error if we try to create another
            # container with the same name.
            with self.assertRaises(subprocess.CalledProcessError):
                with AutograderSandbox(name=self.name):
                    pass

        # The container should have been deleted at this point,
        # so we should be able to create another with the same name.
        with AutograderSandbox(name=self.name):
            pass

    def test_sandbox_environment_variables_set(self):
        print_env_var_script = "echo ${}".format(
            ' $'.join(self.environment_variables))

        sandbox = AutograderSandbox(
            environment_variables=self.environment_variables)
        with sandbox, typing.cast(typing.TextIO,
                                  tempfile.NamedTemporaryFile('w+')) as f:
            f.write(print_env_var_script)
            f.seek(0)
            sandbox.add_files(f.name)
            result = sandbox.run_command(['bash', os.path.basename(f.name)])
            expected_output = ' '.join(
                str(val) for val in self.environment_variables.values())
            expected_output += '\n'
            self.assertEqual(expected_output, result.stdout.read().decode())

    def test_reset(self):
        with AutograderSandbox() as sandbox:
            file_to_add = os.path.abspath(__file__)
            sandbox.add_files(file_to_add)

            ls_result = sandbox.run_command(['ls']).stdout
            self.assertEqual(os.path.basename(file_to_add) + '\n', ls_result.read().decode())

            sandbox.reset()

            ls_result = sandbox.run_command(['ls']).stdout.read().decode()
            self.assertEqual('', ls_result)

    def test_restart_added_files_preserved(self):
        with AutograderSandbox() as sandbox:
            file_to_add = os.path.abspath(__file__)
            sandbox.add_files(file_to_add)

            ls_result = sandbox.run_command(['ls']).stdout.read().decode()
            print(ls_result)
            self.assertEqual(os.path.basename(file_to_add) + '\n', ls_result)

            sandbox.restart()

            ls_result = sandbox.run_command(['ls']).stdout.read().decode()
            self.assertEqual(os.path.basename(file_to_add) + '\n', ls_result)

    def test_entire_process_tree_killed_on_timeout(self):
        for program_str in _PROG_THAT_FORKS, _PROG_WITH_SUBPROCESS_STALL:
            with AutograderSandbox() as sandbox:
                ps_result = sandbox.run_command(['ps', '-aux']).stdout.read().decode()
                print(ps_result)
                num_ps_lines = len(ps_result.split('\n'))
                print(num_ps_lines)

                script_file = _add_string_to_sandbox_as_file(
                    program_str, '.py', sandbox)

                start_time = time.time()
                result = sandbox.run_command(['python3', script_file], timeout=1)
                self.assertTrue(result.timed_out)

                time_elapsed = time.time() - start_time
                self.assertLess(time_elapsed, _SLEEP_TIME // 2,
                                msg='Killing processes took too long')

                ps_result_after_restart = sandbox.run_command(
                    ['ps', '-aux']).stdout.read().decode()
                print(ps_result_after_restart)
                num_ps_lines_after_restart = len(
                    ps_result_after_restart.split('\n'))
                self.assertEqual(num_ps_lines, num_ps_lines_after_restart)

    def test_try_to_change_cmd_runner(self):
        runner_path = '/usr/local/bin/cmd_runner.py'
        with AutograderSandbox() as sandbox:  # type: AutograderSandbox
            # Make sure the file path above is correct
            sandbox.run_command(['cat', runner_path], check=True)
            with self.assertRaises(subprocess.CalledProcessError):
                sandbox.run_command(['touch', runner_path], check=True)

    @mock.patch('subprocess.run')
    @mock.patch('subprocess.check_call')
    def test_container_create_timeout(self, mock_check_call, *args):
        print(mock)
        with AutograderSandbox(debug=True):
            args, kwargs = mock_check_call.call_args
            self.assertIsNone(kwargs['timeout'])

        timeout = 42
        with AutograderSandbox(container_create_timeout=timeout):
            args, kwargs = mock_check_call.call_args
            self.assertEqual(timeout, kwargs['timeout'])


class AutograderSandboxEncodeDecodeIOTestCase(unittest.TestCase):
    def setUp(self):
        self.non_utf = b'\x80 and some other stuff just because\n'
        with self.assertRaises(UnicodeDecodeError):
            self.non_utf.decode()
        self.file_to_print = 'non-utf.txt'
        with open(self.file_to_print, 'wb') as f:
            f.write(self.non_utf)

    def tearDown(self):
        os.remove(self.file_to_print)

    def test_non_unicode_chars_in_normal_output(self):
        with AutograderSandbox() as sandbox:  # type: AutograderSandbox
            sandbox.add_files(self.file_to_print)

            result = sandbox.run_command(['cat', self.file_to_print])
            stdout = result.stdout.read()
            print(stdout)
            self.assertEqual(self.non_utf, stdout)

            result = sandbox.run_command(['bash', '-c', '>&2 cat ' + self.file_to_print])
            stderr = result.stderr.read()
            print(stderr)
            self.assertEqual(self.non_utf, stderr)

    def test_non_unicode_chars_in_output_command_timed_out(self):
        with AutograderSandbox() as sandbox:
            sandbox.add_files(self.file_to_print)

            result = sandbox.run_command(
                ['bash', '-c', 'cat {}; sleep 5'.format(self.file_to_print)],
                timeout=1)
            self.assertTrue(result.timed_out)
            self.assertEqual(self.non_utf, result.stdout.read())

        with AutograderSandbox() as sandbox:
            sandbox.add_files(self.file_to_print)

            result = sandbox.run_command(
                ['bash', '-c', '>&2 cat {}; sleep 5'.format(self.file_to_print)],
                timeout=1)
            self.assertTrue(result.timed_out)
            self.assertEqual(self.non_utf, result.stderr.read())

    def test_non_unicode_chars_in_output_on_process_error(self):
        with AutograderSandbox() as sandbox:
            sandbox.add_files(self.file_to_print)

            with self.assertRaises(subprocess.CalledProcessError) as cm:
                sandbox.run_command(
                    ['bash', '-c', 'cat {}; exit 1'.format(self.file_to_print)],
                    check=True)
            self.assertEqual(self.non_utf, cm.exception.stdout.read())

        with AutograderSandbox() as sandbox:
            sandbox.add_files(self.file_to_print)

            with self.assertRaises(subprocess.CalledProcessError) as cm:
                sandbox.run_command(
                    ['bash', '-c', '>&2 cat {}; exit 1'.format(self.file_to_print)],
                    check=True)
            self.assertEqual(self.non_utf, cm.exception.stderr.read())


_SLEEP_TIME = 6

_PROG_THAT_FORKS = """
import subprocess

print('hello', flush=True)
subprocess.Popen(['sleep', '{}'])
print('goodbye', flush=True)
""".format(_SLEEP_TIME)

_PROG_WITH_SUBPROCESS_STALL = """
import subprocess

print('hello', flush=True)
subprocess.call(['sleep', '{}'])
print('goodbye', flush=True)
""".format(_SLEEP_TIME)


class AutograderSandboxResourceLimitTestCase(unittest.TestCase):

    def setUp(self):
        self.sandbox = AutograderSandbox()

        self.small_virtual_mem_limit = mb_to_bytes(100)
        self.large_virtual_mem_limit = gb_to_bytes(1)

    def test_run_command_timeout_exceeded(self):
        with self.sandbox:
            result = self.sandbox.run_command(["sleep", "10"], timeout=1)
            self.assertTrue(result.timed_out)

    def test_command_exceeds_process_limit(self):
        process_limit = 0
        processes_to_spawn = process_limit + 2

        with self.sandbox:
            self._do_process_resource_limit_test(
                processes_to_spawn, process_limit, self.sandbox)

    def test_command_doesnt_exceed_process_limit(self):
        process_limit = 10
        processes_to_spawn = process_limit - 2

        with self.sandbox:
            self._do_process_resource_limit_test(
                processes_to_spawn, process_limit, self.sandbox)

    def test_command_spawns_no_processes_with_limit_zero(self):
        with self.sandbox:
            self._do_process_resource_limit_test(0, 0, self.sandbox)

    def test_command_exceeds_stack_size_limit(self):
        stack_size_limit = mb_to_bytes(5)
        mem_to_use = stack_size_limit * 2
        with self.sandbox:
            self._do_stack_resource_limit_test(
                mem_to_use, stack_size_limit, self.sandbox)

    def test_command_doesnt_exceed_stack_size_limit(self):
        stack_size_limit = mb_to_bytes(30)
        mem_to_use = stack_size_limit // 2
        with self.sandbox:
            self._do_stack_resource_limit_test(
                mem_to_use, stack_size_limit, self.sandbox)

    def test_command_exceeds_virtual_mem_limit(self):
        virtual_mem_limit = mb_to_bytes(100)
        mem_to_use = virtual_mem_limit * 2
        with self.sandbox:
            self._do_heap_resource_limit_test(
                mem_to_use, virtual_mem_limit, self.sandbox)

    def test_command_doesnt_exceed_virtual_mem_limit(self):
        virtual_mem_limit = mb_to_bytes(100)
        mem_to_use = virtual_mem_limit // 2
        with self.sandbox:
            self._do_heap_resource_limit_test(
                mem_to_use, virtual_mem_limit, self.sandbox)

    def test_run_subsequent_commands_with_different_resource_limits(self):
        with self.sandbox:
            # Under limit
            self._do_stack_resource_limit_test(
                mb_to_bytes(1), mb_to_bytes(10), self.sandbox)
            # Over previous limit
            self._do_stack_resource_limit_test(
                mb_to_bytes(20), mb_to_bytes(10), self.sandbox)
            # Limit raised
            self._do_stack_resource_limit_test(
                mb_to_bytes(20), mb_to_bytes(50), self.sandbox)
            # Over new limit
            self._do_stack_resource_limit_test(
                mb_to_bytes(40), mb_to_bytes(30), self.sandbox)

            # Under limit
            self._do_heap_resource_limit_test(
                mb_to_bytes(10), mb_to_bytes(100), self.sandbox)
            # Over previous limit
            self._do_heap_resource_limit_test(
                mb_to_bytes(200), mb_to_bytes(100), self.sandbox)
            # Limit raised
            self._do_heap_resource_limit_test(
                mb_to_bytes(200), mb_to_bytes(300), self.sandbox)
            # Over new limit
            self._do_heap_resource_limit_test(
                mb_to_bytes(250), mb_to_bytes(200), self.sandbox)

            # Under limit
            self._do_process_resource_limit_test(0, 0, self.sandbox)
            # Over previous limit
            self._do_process_resource_limit_test(2, 0, self.sandbox)
            # Limit raised
            self._do_process_resource_limit_test(2, 5, self.sandbox)
            # Over new limit
            self._do_process_resource_limit_test(10, 7, self.sandbox)

    def _do_stack_resource_limit_test(self, mem_to_use, mem_limit, sandbox):
        prog_ret_code = _run_stack_usage_prog(mem_to_use, mem_limit, sandbox)

        self._check_resource_limit_test_result(
            prog_ret_code, mem_to_use, mem_limit)

    def _do_heap_resource_limit_test(self, mem_to_use, mem_limit, sandbox):
        prog_ret_code = _run_heap_usage_prog(mem_to_use, mem_limit, sandbox)
        self._check_resource_limit_test_result(
            prog_ret_code, mem_to_use, mem_limit)

    def _do_process_resource_limit_test(self, num_processes_to_spawn,
                                        process_limit, sandbox):
        prog_ret_code = _run_process_spawning_prog(
            num_processes_to_spawn, process_limit, sandbox)

        self._check_resource_limit_test_result(
            prog_ret_code, num_processes_to_spawn, process_limit)

    def _check_resource_limit_test_result(self, ret_code, resource_used,
                                          resource_limit):
        if resource_used > resource_limit:
            self.assertNotEqual(0, ret_code)
        else:
            self.assertEqual(0, ret_code)

    def test_multiple_containers_dont_exceed_ulimits(self):
        """
        One quirk of docker containers is that if there are multiple users
        created in different containers but with the same UID, the resource
        usage of all those users will contribute to hitting the same ulimits.
        This test makes sure that valid processes aren't randomly cut off.
        """
        self._do_parallel_container_stack_limit_test(
            16, mb_to_bytes(20), mb_to_bytes(30))

        self._do_parallel_container_heap_limit_test(
            16, mb_to_bytes(300), mb_to_bytes(500))

        self._do_parallel_container_process_limit_test(16, 3, 5)
        self._do_parallel_container_process_limit_test(15, 0, 0)

    def _do_parallel_container_stack_limit_test(self, num_containers,
                                                mem_to_use, mem_limit):
        self._do_parallel_container_resource_limit_test(
            _run_stack_usage_prog, num_containers, mem_to_use, mem_limit)

    def _do_parallel_container_heap_limit_test(self, num_containers,
                                               mem_to_use, mem_limit):
        self._do_parallel_container_resource_limit_test(
            _run_heap_usage_prog, num_containers, mem_to_use, mem_limit)

    def _do_parallel_container_process_limit_test(self, num_containers,
                                                  num_processes_to_spawn,
                                                  process_limit):
        self._do_parallel_container_resource_limit_test(
            _run_process_spawning_prog, num_containers,
            num_processes_to_spawn, process_limit)

    def _do_parallel_container_resource_limit_test(self, func_to_run,
                                                   num_containers,
                                                   amount_to_use,
                                                   resource_limit):
        with multiprocessing.Pool(processes=num_containers) as p:
            return_codes = p.starmap(
                func_to_run,
                itertools.repeat((amount_to_use, resource_limit, None),
                                 num_containers))

        print(return_codes)
        for ret_code in return_codes:
            self.assertEqual(0, ret_code)


def _run_stack_usage_prog(mem_to_use, mem_limit, sandbox):
    def _run_prog(sandbox):
        prog = _STACK_USAGE_PROG_TMPL.format(num_bytes_on_stack=mem_to_use)
        filename = _add_string_to_sandbox_as_file(prog, '.cpp', sandbox)
        exe_name = _compile_in_sandbox(sandbox, filename)
        result = sandbox.run_command(
            ['./' + exe_name], max_stack_size=mem_limit)
        return result.return_code

    return _call_function_and_allocate_sandbox_if_needed(_run_prog, sandbox)


_STACK_USAGE_PROG_TMPL = """#include <iostream>
#include <thread>

using namespace std;

int main()
{{
    char stacky[{num_bytes_on_stack}];

    this_thread::sleep_for(chrono::seconds(2));

    cout << stacky << endl;
    return 0;
}}
"""


def _run_heap_usage_prog(mem_to_use, mem_limit, sandbox):
    def _run_prog(sandbox):
        prog = _HEAP_USAGE_PROG_TMPL.format(num_bytes_on_heap=mem_to_use)
        filename = _add_string_to_sandbox_as_file(prog, '.cpp', sandbox)
        exe_name = _compile_in_sandbox(sandbox, filename)
        result = result = sandbox.run_command(
            ['./' + exe_name], max_virtual_memory=mem_limit)

        return result.return_code

    return _call_function_and_allocate_sandbox_if_needed(_run_prog, sandbox)


_HEAP_USAGE_PROG_TMPL = """#include <iostream>
#include <thread>

using namespace std;

int main()
{{
    char* heapy = new char[{num_bytes_on_heap}];

    this_thread::sleep_for(chrono::seconds(2));

    cout << heapy << endl;
    return 0;
}}
"""


def _compile_in_sandbox(sandbox, *files_to_compile):
    exe_name = 'prog'
    sandbox.run_command(
        ['g++', '--std=c++11'] + list(files_to_compile) +
        ['-o', exe_name], check=True)
    return exe_name


def _run_process_spawning_prog(num_processes_to_spawn, process_limit,
                               sandbox):
    def _run_prog(sandbox):
        prog = _PROCESS_SPAWN_PROG_TMPL.format(
            num_processes=num_processes_to_spawn)
        filename = _add_string_to_sandbox_as_file(prog, '.py', sandbox)

        result = sandbox.run_command(['python3', filename],
                                     max_num_processes=process_limit)
        return result.return_code

    return _call_function_and_allocate_sandbox_if_needed(_run_prog, sandbox)


_PROCESS_SPAWN_PROG_TMPL = """
import time
import subprocess


processes = []
for i in range({num_processes}):
    proc = subprocess.Popen(['sleep', '2'])
    processes.append(proc)

time.sleep(2)

for proc in processes:
    proc.communicate()
"""


def _add_string_to_sandbox_as_file(string, file_extension, sandbox):
    with tempfile.NamedTemporaryFile('w+', suffix=file_extension) as f:
        f.write(string)
        f.seek(0)
        sandbox.add_files(f.name)

        return os.path.basename(f.name)


def _call_function_and_allocate_sandbox_if_needed(func, sandbox):
    if sandbox is None:
        sandbox = AutograderSandbox()
        with sandbox:
            return func(sandbox)
    else:
        return func(sandbox)


# -----------------------------------------------------------------------------


_GOOGLE_IP_ADDR = "216.58.214.196"


class AutograderSandboxNetworkAccessTestCase(unittest.TestCase):

    def setUp(self):
        super().setUp()

        self.google_ping_cmd = ['ping', '-c', '5', _GOOGLE_IP_ADDR]

    def test_networking_disabled(self):
        with AutograderSandbox() as sandbox:
            result = sandbox.run_command(self.google_ping_cmd)
            self.assertNotEqual(0, result.return_code)

    def test_networking_enabled(self):
        with AutograderSandbox(allow_network_access=True) as sandbox:
            result = sandbox.run_command(self.google_ping_cmd)
            self.assertEqual(0, result.return_code)

    def test_set_allow_network_access(self):
        sandbox = AutograderSandbox()
        self.assertFalse(sandbox.allow_network_access)
        with sandbox:
            result = sandbox.run_command(self.google_ping_cmd)
            self.assertNotEqual(0, result.return_code)

        sandbox.allow_network_access = True
        self.assertTrue(sandbox.allow_network_access)
        with sandbox:
            result = sandbox.run_command(self.google_ping_cmd)
            self.assertEqual(0, result.return_code)

        sandbox.allow_network_access = False
        self.assertFalse(sandbox.allow_network_access)
        with sandbox:
            result = sandbox.run_command(self.google_ping_cmd)
            self.assertNotEqual(0, result.return_code)

    def test_error_set_allow_network_access_while_running(self):
        with AutograderSandbox() as sandbox:
            with self.assertRaises(ValueError):
                sandbox.allow_network_access = True

            self.assertFalse(sandbox.allow_network_access)
            result = sandbox.run_command(self.google_ping_cmd)
            self.assertNotEqual(0, result.return_code)


class AutograderSandboxCopyFilesTestCase(unittest.TestCase):

    def test_copy_files_into_sandbox(self):
        files = []
        try:
            for i in range(10):
                f = tempfile.NamedTemporaryFile(mode='w+')
                f.write('this is file {}'.format(i))
                f.seek(0)
                files.append(f)

            filenames = [file_.name for file_ in files]

            with AutograderSandbox() as sandbox:
                sandbox.add_files(*filenames)

                ls_result = sandbox.run_command(['ls']).stdout.read().decode()
                actual_filenames = [
                    filename.strip() for filename in ls_result.split()]
                expected_filenames = [
                    os.path.basename(filename) for filename in filenames]
                self.assertCountEqual(expected_filenames, actual_filenames)

                for file_ in files:
                    file_.seek(0)
                    expected_content = file_.read()
                    actual_content = sandbox.run_command(
                        ['cat', os.path.basename(file_.name)]
                    ).stdout.read().decode()
                    self.assertEqual(expected_content, actual_content)
        finally:
            for file_ in files:
                file_.close()

    def test_copy_and_rename_file_into_sandbox(self):
        expected_content = 'this is a file'
        with tempfile.NamedTemporaryFile('w+') as f:
            f.write(expected_content)
            f.seek(0)

            with AutograderSandbox() as sandbox:
                new_name = 'new_filename.txt'
                sandbox.add_and_rename_file(f.name, new_name)

                ls_result = sandbox.run_command(['ls']).stdout.read().decode()
                actual_filenames = [filename.strip() for filename in ls_result.split()]
                expected_filenames = [new_name]
                self.assertCountEqual(expected_filenames, actual_filenames)

                actual_content = sandbox.run_command(['cat', new_name]).stdout.read().decode()
                self.assertEqual(expected_content, actual_content)

    def test_add_files_root_owner_and_read_only(self):
        original_content = "some stuff you shouldn't change"
        overwrite_content = 'lol I changed it anyway u nub'
        with tempfile.NamedTemporaryFile('w+') as f:
            f.write(original_content)
            f.seek(0)

            added_filename = os.path.basename(f.name)

            with AutograderSandbox() as sandbox:
                sandbox.add_files(f.name, owner='root', read_only=True)

                actual_content = sandbox.run_command(
                    ['cat', added_filename], check=True
                ).stdout.read().decode()
                self.assertEqual(original_content, actual_content)

                with self.assertRaises(subprocess.CalledProcessError):
                    sandbox.run_command(['touch', added_filename], check=True)

                with self.assertRaises(subprocess.CalledProcessError):
                    sandbox.run_command(
                        ['bash', '-c',
                         "printf '{}' > {}".format(overwrite_content, added_filename)],
                        check=True)

                actual_content = sandbox.run_command(
                    ['cat', added_filename], check=True
                ).stdout.read().decode()
                self.assertEqual(original_content, actual_content)

                root_touch_result = sandbox.run_command(
                    ['touch', added_filename], check=True, as_root=True)
                self.assertEqual(0, root_touch_result.return_code)

                sandbox.run_command(
                    ['bash', '-c', "printf '{}' > {}".format(overwrite_content, added_filename)],
                    as_root=True, check=True)
                actual_content = sandbox.run_command(
                    ['cat', added_filename]
                ).stdout.read().decode()
                self.assertEqual(overwrite_content, actual_content)

    def test_overwrite_non_read_only_file(self):
        original_content = "some stuff"
        overwrite_content = 'some new stuff'
        with tempfile.NamedTemporaryFile('w+') as f:
            f.write(original_content)
            f.seek(0)

            added_filename = os.path.basename(f.name)

            with AutograderSandbox() as sandbox:
                sandbox.add_files(f.name)

                actual_content = sandbox.run_command(
                    ['cat', added_filename], check=True
                ).stdout.read().decode()
                self.assertEqual(original_content, actual_content)

                sandbox.run_command(
                    ['bash', '-c', "printf '{}' > {}".format(overwrite_content, added_filename)])
                actual_content = sandbox.run_command(
                    ['cat', added_filename], check=True
                ).stdout.read().decode()
                self.assertEqual(overwrite_content, actual_content)

    def test_error_add_files_invalid_owner(self):
        with AutograderSandbox() as sandbox:
            with self.assertRaises(ValueError):
                sandbox.add_files('steve', owner='not_an_owner')

if __name__ == '__main__':
    unittest.main()
